unit fb01;   {Version 2.0}

{ Programming unit for the Yamaha FB-01 (and IBM Music Feature)
  Copyright (C) 1989, 1990,  Kevin Weiner
 }

INTERFACE

const

{ Instrument parameter offsets }

  inst_notes    = $00;
  inst_chan     = $01;
  inst_high     = $02;
  inst_low      = $03;
  inst_bank     = $04;
  inst_voice    = $05;
  inst_detune   = $06;
  inst_oct      = $07;
  inst_output   = $08;
  inst_pan      = $09;
  inst_lfoena   = $0a;
  inst_port     = $0b;
  inst_pbend    = $0c;
  inst_mode     = $0d;
  inst_pmdctrl  = $0e;
  inst_lfospeed = $10;
  inst_amd      = $11;
  inst_pmd      = $12;
  inst_lfowave  = $13;
  inst_lfoload  = $14;
  inst_lfosync  = $15;
  inst_ams      = $16;
  inst_pms      = $17;

{ Voice parameter offsets      data masks           mask postition
                                                    (shift left)    }
  vp_name       = $00;
  vp_lfospeed   = $08;     vm_lfospeed  = 255;     vs_lfospeed  = 0;
  vp_lfoload    = $09;     vm_lfoload   =   1;     vs_lfoload   = 7;
  vp_amd        = $09;     vm_amd       = 127;     vs_amd       = 0;
  vp_lfosync    = $0a;     vm_lfosync   =   1;     vs_lfosync   = 7;
  vp_pmd        = $0a;     vm_pmd       = 127;     vs_pmd       = 0;
  vp_opena4     = $0b;     vm_opena4    =   1;     vs_opena4    = 6;
  vp_opena3     = $0b;     vm_opena3    =   1;     vs_opena3    = 5;
  vp_opena2     = $0b;     vm_opena2    =   1;     vs_opena2    = 4;
  vp_opena1     = $0b;     vm_opena1    =   1;     vs_opena1    = 3;
  vp_fblevel    = $0c;     vm_fblevel   =   7;     vs_fblevel   = 3;
  vp_alg        = $0c;     vm_alg       =   7;     vs_alg       = 0;
  vp_pms        = $0d;     vm_pms       =   7;     vs_pms       = 4;
  vp_ams        = $0d;     vm_ams       =   3;     vs_ams       = 0;
  vp_lfowave    = $0e;     vm_lfowave   =   3;     vs_lfowave   = 5;
  vp_transp     = $0f;     vm_transp    = 255;     vs_transp    = 0;

  vp_op1        = $10;  {operator offsets}
  vp_op2        = $18;
  vp_op3        = $20;
  vp_op4        = $28;

  vp_mode       = $3a;     vm_mode      =   1;     vs_mode      = 7;
  vp_port       = $3a;     vm_port      = 127;     vs_port      = 0;
  vp_pmdctrl    = $3b;     vm_pmdctrl   =   7;     vs_pmdctrl   = 4;
  vp_pbend      = $3b;     vm_pbend     =  15;     vs_pbend     = 0;

{ Individual operator parameters }

  vo_tlevel     = $00;     vm_tlevel    = 127;     vs_tlevel    = 0;
  vo_kls0       = $01;     vm_kls0      =   1;     vs_kls0      = 7;
  vo_vstl       = $01;     vm_vstl      =   7;     vs_vstl      = 4;
  vo_klsd       = $02;     vm_klsd      =  15;     vs_klsd      = 4;
  vo_adjtl      = $02;     vm_adjtl     =  15;     vs_adjtl     = 0;
  vo_kls1       = $03;     vm_kls1      =   1;     vs_kls1      = 7;
  vo_detune     = $03;     vm_detune    =   7;     vs_detune    = 4;
  vo_mult       = $03;     vm_mult      =  15;     vs_mult      = 0;
  vo_krsd       = $04;     vm_krsd      =   3;     vs_krsd      = 6;
  vo_attack     = $04;     vm_attack    =  31;     vs_attack    = 0;
  vo_type       = $05;     vm_type      =   1;     vs_type      = 7;
  vo_vsar       = $05;     vm_vsar      =   3;     vs_vsar      = 5;
  vo_decay1     = $05;     vm_decay1    =  31;     vs_decay1    = 0;
  vo_inharm     = $06;     vm_inharm    =   3;     vs_inharm    = 6;
  vo_decay2     = $06;     vm_decay2    =  31;     vs_decay2    = 0;
  vo_sust       = $07;     vm_sust      =  15;     vs_sust      = 4;
  vo_rel        = $07;     vm_rel       =  15;     vs_rel       = 0;

type

{ One configuration data set }
  configdata = array [0..159] of byte;

{ One voice data set }
  voicedata = array [0..63] of byte;

{ All voices in one bank (-1 is header record) }
  bankdata = array [-1..47] of voicedata;

{ Packet type selection: a = 8-bit data,  b = 7-bit data }
  packettype = (a_packet, b_packet);

var
  lastack: boolean;                    {Last dump ack result}

{ FB01 user routines --------------------------------------------------------}


procedure getpacket(p: packettype);    {Write a data packet to the synth}

procedure putpacket(p: packettype);    {Write a data packet to the synth}

procedure setnodeparam(param, data: integer);    {Set a common node param}

procedure setinstparam(inst, param, data: integer);    {Set 7-bit instrument
                                                        parameter (config)}

procedure setvoiceparam(inst, param, data: integer);   {Set 8-bit instrument
                                                        parameter (voice)}

procedure getconfig(n: integer; var data: configdata; var err: boolean);
                                       {Read data for specified config}

procedure putconfig(n: integer; var data: configdata);   {Write config}

procedure putvoice(inst: integer; var data: voicedata);  {Write voice data for
                                                          selected instrument}

procedure getbank(n: integer; var data: bankdata; var err: boolean);
                                       {Read data for specified voice bank}

procedure putbank(n: integer; var data: bankdata);    {Write bank}

procedure storeconfig(n: integer);     {Store current configuration buffer}

procedure storevoice(inst, bank, voice: integer);     {Store current instrument
                                                       voice to bank}
procedure nodedump(source, id: integer);  {Request dump of current node}

function checknode: boolean;           {Check current node id}


IMPLEMENTATION

uses crt, dos, mfc;

const

{Misc. constants}
  retries = 2;                         {Retry limit on packet errors}
  hexdigits: string[16] = '0123456789ABCDEF';


function hex(n: integer): string;

  { Convert integer to hex string }

  var
    s: string;

  begin
    if n < 0 then
      hex := '-1'
    else
      begin
        s := '';
        repeat
          s := hexdigits[(n mod 16)+1] + s;
          n := n div 16;
        until n = 0;
        hex := s;
      end;
  end;

procedure sysex;

  {Write 4-byte system exclusive header}

  begin
    putbyte($f0);  putbyte($43);  putbyte($75);  putbyte(node);
  end;


procedure nodedump { source, id: integer };

  {Request dump of current node from specified storage source/id}

  var x: integer;

  begin
    if mdebug then write(' nodedump:',hex(source),':',hex(id));
    sysex;
    putbyte($20);
    putbyte(source);
    putbyte(id);
    putbyte($f7);
  end;

procedure getpacket { p: packettype };

 {Read a packet of the specified type (a_packet=8-bit, b_packet=7-bit).
  Stores packet to packet_buffer with size packet_size.  Packet_error
  indicates read failure if true.}

  var
    x, x1, y, low, size, chksum, testsum: integer;

  begin
    if mdebug then write(' getpacket');
    packeterror := true;
    testsum := 0;
    getbyte(size);                     {get 2-byte packet size}
    if lastdata >= 0 then
      begin
        getbyte(low);
        size := size * 128 + low;
        if p = a_packet then size := size div 2;
        if mdebug then write(' size=',size);
        if size > 0 then
          begin
            packetsize := 0;
            repeat
              getbyte(x1);
              x := x1;
              testsum := testsum + x;      {Add to checksum}
              if p = a_packet then
                begin
                  getbyte(y);              {2nd byte if 8-bit data}
                  testsum := testsum + y;
                  x := y * 16 + x;
                end;
              packetsize := packetsize + 1;
              if (x < 0) or (x > 256) then x := 0;
              packetbuffer[packetsize] := x;
            until (x1 < 0) or (packetsize = size);
            if mdebug and (x1 < 0) then write(' packet timeout');

            if x1 >= 0 then               {Compare checksums}
              begin
                getbyte(chksum);
                packeterror := ((-testsum) and 127) <> chksum;
                if mdebug and packeterror then write(' checksum error');
              end;
            if mdebug and not packeterror then write(' good packet');
          end;
      end
    else
      if mdebug then write(' no packet');
  end;

function ack: boolean;

  {Wait for status message returned after dump to node.
   Ack true indicates success.}

  var x: integer;

  begin
    if mdebug then write( 'wait ack-');
    waitbyte($f0);
    if lastdata <> $f0 then
      begin
        if waitack then
          lastack := false
        else
          lastack := true;
      end
    else
      begin
        waitbyte($60+node);
        getbyte(x);
        lastack := x = 2;
        getbyte(x);
      end;
    ack := lastack;
    if mdebug then write(lastack);
  end;

procedure putpacket { p: packettype };

  {Write a packet of specified type (a_packet=8-bit, b_packet=7-bit).
   Assumes packet_size bytes in packet_buffer.}

  var
    i, x, y, chksum: integer;

  begin
    if mdebug then write(' putpacket');
    i := packetsize;
    if p = a_packet then i := i * 2;
    putbyte(i div 128);                {2-byte size}
    putbyte(i and 127);

    chksum := 0;
    for i := 1 to packetsize do
      begin
        x := packetbuffer[i];
        if p = a_packet then
          begin
            y := x and 15;
            putbyte(y);                {First byte if 8-bit data}
            chksum := chksum + y;
            x := x div 16;
          end;
        putbyte(x);                    {7-bit data byte, or second byte
                                        if 8-bit data}
        chksum := chksum + x;
      end;
    putbyte((-chksum) and 127);        {Write checksum}
  end;

procedure setnodeparam { param, data: integer };

  {Set value of specified node parameter}

  begin
    if mdebug then write(' nodeparam:',hex(param),':',hex(data));
    if data < 0 then
      data := data+128;                {7-bit 2's comp if negative}
    sysex;
    putbyte($10);
    putbyte(param);
    putbyte(data);
    putbyte($f7);
    if mfcmode and (param >= node_number) then    {Keep track of changes}
      nstat[currentmfc,param-node_number] := data;
  end;

procedure setinstparam { inst, param, data: integer };

  {Set value of specified instrument configuration parameter (7-bit)}

  begin
    if mdebug then write(' instparam:',hex(inst),':',hex(param),':',hex(data));
    if data < 0 then
      data := data+128;                {7-bit 2's comp if negative}
    sysex;
    putbyte($18+inst);
    putbyte(param);
    putbyte(data);
    putbyte($f7);
  end;

procedure setvoiceparam { inst, param, data: integer };

  {Set value of specified instrument voice parameter (8-bit)}

  begin
    if mdebug then write(' voiceparam:',hex(inst),':',hex(param),':',hex(data));
    sysex;
    putbyte($18+inst);
    putbyte(param+$40);                {Offset of $40 indicates voice data}
    putbyte(data and $f);              {Low}
    putbyte(data shr 4);               {High}
    putbyte($f7);
  end;

procedure getconfig { n: integer; var data: configdata; var err: boolean };

  {Read configuration number n.  -1 = current configuration buffer.
   err returned true if packet error on read; packet_error also set.}

  var
    i, x, tries, src: integer;

  begin
    if mdebug then write(' getconfig');
    if n < 0 then
      begin src := 1; n := 0; end      {Select current config}
    else
      src := 2;                        {or one of stored configs}
    tries := 0;
    repeat
      nodedump(src, n);
      waitbyte($75);
      if lastdata = $75 then
        begin
          for i := 1 to 3 do getbyte(x);
          waitbyte(n);
          getpacket(b_packet);
        end
      else
        packeterror := true;
      tries := tries + 1;
    until (not packeterror) or (tries > retries);
    waitbyte($F7);
    if packeterror or (packetsize <> 160) then
      err := true
    else
      begin
        err := false;                  {Successful - copy data from buffer}
        for i := 1 to packetsize do
          data[i-1] := packetbuffer[i];
      end;
  end;

procedure putconfig { n: integer; var data: configdata };

  {Write configuration data to specified config store.  packet_error
   set if error writing packet}

  var
    i, tries: integer;

  begin
    if mdebug then write(' putconfig');
    for i := 1 to 160 do
      packetbuffer[i] := data[i-1];
    packetsize := 160;
    tries := 0;
    repeat
      sysex;
      putbyte(0);
      putbyte(2);
      putbyte(n);
      putpacket(b_packet);
      putbyte($f7);
      packeterror := not ack;
      tries := tries + 1;
    until (not packeterror) or (tries > retries);
  end;

procedure putvoice { inst: integer; var data: voicedata };

  {Write voice parameter data to specified instrument}

  var
    i, tries: integer;

  begin
    if mdebug then write(' putvoice');
    for i := 1 to 64 do
      packetbuffer[i] := data[i-1];
    packetsize := 64;
    tries := 0;
    repeat
      sysex;
      putbyte(8+inst);
      putbyte(0);
      putbyte(0);
      putpacket(a_packet);
      putbyte($f7);
      packeterror := not ack;
      tries := tries + 1;
    until (not packeterror) or (tries > retries);
  end;


procedure getbank { n: integer; var data: bankdata; var err: boolean };

  {Read specified voice bank.  err true if packet error}

  var
    i, x, tries, vnum: integer;

  begin
    if mdebug then write(' getbank');
    tries := 0;
    repeat
      nodedump(0, n);
      waitbyte($75);
      if lastdata = $75 then
        begin
          for i := 1 to 3 do getbyte(x);
          waitbyte(n);
          vnum := -1;
          repeat
            if mdebug then write(' packet ',vnum);
            getpacket(a_packet);
            if not packeterror then
              for i := 1 to 64 do
                data[vnum,i-1] := packetbuffer[i];
            vnum := vnum + 1;
          until packeterror or (vnum > 47);
          tries := tries + 1;
          if packeterror then waitbyte(-1) else waitbyte($f7);
        end
      else
        begin
          packeterror := true;
          tries := tries + 1;
        end;
    until (not packeterror) or (tries > retries);
    err := packeterror;
  end;

procedure putbank { n: integer; var data: bankdata };

  {Write specified voice bank}

  var
    i, x, tries, vnum: integer;

  begin
    if mdebug then write(' putbank');
    tries := 0;
    repeat
      sysex;
      putbyte(0);
      putbyte(0);
      putbyte(n and 1);
      for vnum := -1 to 47 do
        begin
          if vnum < 0 then
            packetsize := 32  {header size}
          else
            packetsize := 64; {data size}
          for i := 1 to packetsize do
            packetbuffer[i] := data[vnum,i-1];
          putpacket(a_packet);
        end;
      putbyte($f7);
      tries := tries + 1;
    until ack or (tries > retries);
  end;

procedure storeconfig { n: integer };

  {Move current config buffer contents into config store n}

  var x: integer;

  begin
    if mdebug then write(' storeconfig');
    sysex;
    putbyte($20);
    putbyte($40);
    putbyte(n);
    putbyte($f7);
    waitbyte($f7);
  end;

procedure storevoice { inst, bank, voice: integer };

  {Move current instrument voice data into specified bank/voice}

  var x: integer;

  begin
    if mdebug then write(' storevoice');
    sysex;
    putbyte($28+inst);
    putbyte($40);
    putbyte((bank and 1)*48+voice);
    putbyte($f7);
    waitbyte($f7);
  end;

function checknode: boolean;

  {Check whether current node will answer requests and verify identification
   string.  Returns true if ok.}

  var
    x: integer;
    ok: boolean;

  begin
    if mdebug then write(' checknode:',node,':');
    nodedump(4, 0);
    waitbyte($f0);
    ok := lastdata > 0;
    if ok then waitbyte($75);
    ok := lastdata > 0;
    if ok then waitbyte(node);
    ok := lastdata >= 0;
    if ok then
      begin
        getbyte(x); getbyte(x); getbyte(x);
        getpacket(b_packet);
        getbyte(x);
        ok := (not packeterror) and
              (packetbuffer[1] = ord('Y')) and
              (packetbuffer[2] = ord('A'));
      end;
    checknode := ok;
    if mdebug then write(ok);
  end;

end.

