/* Sample Turbo C program demonstrating Music Feature card access.

   Kevin Weiner -- 2/11/90

*/

#include <dos.h>
#include <sys\timeb.h>

int            base, lastdata, piu0, piu1, piu2, pcr, tcr;
unsigned char  txrdy, rxrdy, exr8, ext8;

long mtime(void);
int getbyte(void);
void mfcinit(void);
void note(int chan, int notenum, int vel);
void putbyte(int x);
void setcmd(void);
void setdata(void);
void setpath(int p1, int p2, int p3, int p4, int p5);
void waitbyte(int data);


/* mtime -- Return system time in milliseconds */

long mtime(void)
{
    struct timeb timer;

    ftime(&timer);
    return(timer.time*1000+timer.millitm);
}

/* getbyte -- Return a data byte from the MFC, or -1 if none */

int getbyte(void)
{
    long time, starttime;
    unsigned char stat;

    time = 0;
    stat = inportb(piu2);              /* Check status */
    if ((stat & rxrdy) == 0) {         /* No data - wait */
      starttime = mtime();
      while (((stat & rxrdy) == 0) & (time < 500)) {
	 stat = inportb(piu2);
	 time = mtime() - starttime;
      }
    }
    lastdata = -1;
    if (time < 500) {
      lastdata = inportb(piu0);        /* Read data */
      if ((stat & exr8) != 0)
        lastdata = lastdata + 256;     /* Add 256 if command bit set */
    }
    return(lastdata);
}

/* mfcinit -- Initialize MFC parameters, set paths */

void mfcinit(void)
{
    base = 0x2A20;                     /* Port address for 1st mfc */
    lastdata = -1;
    piu0 = base + 0;                   /* Parallel i/o registers */
    piu1 = base + 1;
    piu2 = base + 2;
    pcr = base + 3;                    /* PIU control register */
    tcr = base + 8;                    /* Total control register */
    txrdy = 2;                         /* Ok to send */
    rxrdy = 0x20;                      /* Data ready to read */
    exr8 = 0x80;                       /* 8th bit read */
    ext8 = 0x10;                       /* Transmit 8th bit */
    if (inportb(piu2) == 0xff)
      printf("Music Feature Card not present\n");
    else {
      outportb(pcr, 0xbc);
      setpath(0x1f, 0x1f, 0x1f, 0x1f, 0x1f);
    }
}

/* note -- Play note on channel (1-16) with specified velocity */

void note(int chan, int notenum, int vel)
{
    putbyte(0x90 + chan - 1);
    putbyte(notenum);
    putbyte(vel);
}

/* putbyte -- Write a data byte to the MFC */

void putbyte(int x)
{
    long time, starttime;
    unsigned char stat;

    time = 0;
    stat = inportb(piu2);              /* Check status */
    if ((stat & txrdy) == 0) {         /* Not ready - wait */
      starttime = mtime();
      while (((stat & txrdy) == 0) & (time < 500)) {
	 stat = inportb(piu2);
	 time = mtime() - starttime;
      }
    }
    outportb(piu1, (x & 0xff));
}

/* setcmd -- Set MFC to command mode for subsequent output */

void setcmd(void)
{
    outportb(tcr, ext8);
}

/* setdata -- Set MFC to data mode for subsequent output */

void setdata(void)
{
    outportb(tcr, 0);
}

/* setpath -- Set MFC MIDI paths

   p1 - MIDI IN to System
   p2 - System to MIDI OUT
   p3 - MIDI IN to Sound Processor
   p4 - System to Sound Processor
   p5 - MIDI IN to MIDI OUT
   (Set to 0x1f to enable, 0 to disable)
*/

void setpath(int p1, int p2, int p3, int p4, int p5)
{
    setcmd();
    putbyte(0x1e2);
    putbyte(p1);  putbyte(p2);  putbyte(p3);  putbyte(p4);  putbyte(p5);
    waitbyte(0x1e2);
    setdata();
}

/* waitbyte -- Wait for specified byte from MFC, or timeout */

void waitbyte(int data)
{
    int x;

    x = getbyte();
    while ((x != data) & (x >= 0)) x = getbyte();
}

main()
{
  int x;

  mfcinit();

  note(1, 64, 64);                     /* Play a note */
  delay(1000);
  note(1, 64, 0);

  /* Echo in to out until key pressed */

  printf("Waiting for MIDI data - press a key to quit\n");

  setpath(0x1f, 0x1f, 0, 0x1f, 0);     /* Disable MFC thru paths */

  while (kbhit() == 0) {
    x = getbyte();
    if (x >= 0) { putbyte(x);  printf("%d ",x); };
  }
  printf("\n");
}
