unit queue;

{ MIDI queue handler for unit MFC }

INTERFACE

const
  mmax = $3fff;   {16383}

var
  queued: integer;

procedure dequeue(var x: integer);     {Get next item from queue}
procedure enqueue(x: integer);         {Add next item to queue}
procedure push(x: byte);               {Put item back in queue}
procedure initqueue;                   {Initialize queue to empty}

IMPLEMENTATION

var
  q: array [0..mmax] of byte;
  inptr, outptr: integer;
  stuff: boolean;

procedure dequeue(var x: integer);

  begin
    if queued <= 0 then
      begin
        x := -1;
        queued := 0;
        outptr := inptr;
      end
    else
      begin
        x := q[outptr];
        outptr := (outptr + 1) and mmax;
        queued := queued - 1;
        if x = $ff then                {Special byte stuffing}
          begin
            x := q[outptr];
            outptr := (outptr + 1) and mmax;
            queued := queued - 1;
            if x <> $ff then           {High order byte}
              begin
                x := (x shl 8) + q[outptr];
                outptr := (outptr + 1) and mmax;
                queued := queued - 1;
              end;
          end;
      end;
  end;

procedure enqueue(x: integer);

  var
    newin: integer;
    y: byte;

  begin
    y := byte(x shr 8);                {High order byte}
    if y <> 0 then
      begin                            {Store $ff, hi, low}
        stuff := false;
        enqueue($ff);
        enqueue(y and $7f);
        enqueue(byte(x));
        stuff := true;
      end
    else if (x = $ff) and stuff then   {Store $ff, $ff}
      begin
        stuff := false;
        enqueue($ff);
        enqueue($ff);
        stuff := true;
      end
    else
      begin
        newin := (inptr+1) and mmax;
        if newin <> outptr then
          begin
            q[inptr] := x;
            inptr := newin;
            queued := queued + 1;
          end;
      end;
  end;

procedure push(x: byte);

  begin
    outptr := outptr - 1;
    if outptr < 0 then
      outptr := mmax;
    q[outptr] := x;
    queued := queued + 1;
  end;

procedure initqueue;

  begin
    inptr := 0;
    outptr := 0;
    queued := 0;
    stuff := true;
  end;

begin
  initqueue;
end.